"""
We have here the object representing a task.

Tasks contain a reference to the functionality it provides (in ``harpoon.actions``)
as well as options that are used to override those in the image it's attached
to.
"""

from harpoon.errors import BadOption

from input_algorithms.dictobj import dictobj
from option_merge import MergedOptions

class Task(dictobj):
    """
    Used to add extra options associated with the task and to start the action
    from ``harpoon.actions``.

    Also responsible for complaining if the specified action doesn't exist.

    Will also ask the image to complain about any missing environment variables.
    """
    fields = {
          ("action", "run"): "The action to run with this image"
        , ("options", None): "The options to merge with the image options"
        , ("overrides", None): "The options to merge with the root configuration"
        , ("description", ""): "The description of the task"
        , ("label", "Project"): "The namespace when listing tasks"
        }

    def setup(self, *args, **kwargs):
        super(Task, self).setup(*args, **kwargs)
        self.set_description()

    def set_description(self, available_actions=None):
        if not self.description:
            if not available_actions:
                from harpoon.actions import available_actions
            if self.action in available_actions:
                self.description = available_actions[self.action].__doc__

    def run(self, collector, image, available_actions, tasks, **extras):
        """Run this task"""
        task_func = available_actions[self.action]
        configuration = collector.configuration.wrapped()

        if self.options:
            if image:
                configuration.update({"images": {image: self.options}})
            else:
                configuration.update(self.options)

        # args like --port and the like should override what's in the options
        # But themselves be overridden by the overrides
        configuration.update(configuration["args_dict"].as_dict(), source="<args_dict>")

        if self.overrides:
            overrides = {}
            for key, val in self.overrides.items():
                overrides[key] = val
                if isinstance(val, MergedOptions):
                    overrides[key] = dict(val.items())
            configuration.update(overrides)

        if task_func.needs_image:
            self.find_image(image, configuration)
            image = configuration["images"][image]
            image.find_missing_env()

        from harpoon.collector import Collector
        new_collector = Collector()
        new_collector.configuration = configuration
        new_collector.configuration_file = collector.configuration_file
        return task_func(new_collector, image=image, tasks=tasks, **extras)

    def find_image(self, image, configuration):
        """Complain if we don't have an image"""
        images = configuration["images"]
        available = list(images.keys())

        if not image:
            info = {}
            if available:
                info["available"] = available
            raise BadOption("Please use --image to specify an image", **info)

        if image not in images:
            raise BadOption("No such image", wanted=image, available=available)




"""
The functionality itself for each task.

Each task is specified with the ``a_task`` decorator and indicates whether it's
necessary to provide the task with the object containing all the images and/or
one specific image object.
"""
from harpoon.ship.builder import Builder
from harpoon.ship.syncer import Syncer
from harpoon.errors import BadOption

from docker.errors import APIError as DockerAPIError
from textwrap import dedent
import itertools
import logging

log = logging.getLogger("harpoon.actions")

info = {"is_default": True}
default_actions = []
available_actions = {}

class an_action(object):
    """Records a task in the ``available_actions`` dictionary"""
    def __init__(self, needs_image=False):
        self.needs_image = needs_image

    def __call__(self, func):
        available_actions[func.__name__] = func
        func.needs_image = self.needs_image
        if info["is_default"]:
            default_actions.append(func.__name__)
        return func

@an_action(needs_image=True)
def push(collector, image, **kwargs):
    """Push an image"""
    if not image.image_index:
        raise BadOption("The chosen image does not have a image_index configuration", wanted=image.name)
    Builder().make_image(image, collector.configuration["images"], pushing=True)
    Syncer().push(image)

@an_action()
def push_all(collector, **kwargs):
    """Push all the images"""
    configuration = collector.configuration
    configuration["harpoon"].do_push = True
    configuration["harpoon"].only_pushable = True
    make_all(collector, **kwargs)

@an_action(needs_image=True)
def pull(collector, image, **kwargs):
    """Pull an image"""
    if not image.image_index:
        raise BadOption("The chosen image does not have a image_index configuration", wanted=image.name)
    Syncer().pull(image, ignore_missing=image.harpoon.ignore_missing)

@an_action()
def pull_all(collector, **kwargs):
    """Pull all the images"""
    images = collector.configuration["images"]
    for layer in Builder().layered(images, only_pushable=True):
        for image_name, image in layer:
            log.info("Pulling %s", image_name)
            Syncer().pull(image, ignore_missing=image.harpoon.ignore_missing)

@an_action(needs_image=True)
def make(collector, image, **kwargs):
    """Just create an image"""
    Builder().make_image(image, collector.configuration["images"])
    print("Created image {0}".format(image.image_name))

@an_action()
def make_all(collector, **kwargs):
    """Creates all the images in layered order"""
    configuration = collector.configuration
    push = configuration["harpoon"].do_push
    only_pushable = configuration["harpoon"].only_pushable
    if push:
        only_pushable = True

    images = configuration["images"]
    for layer in Builder().layered(images, only_pushable=only_pushable):
        for _, image in layer:
            Builder().make_image(image, images, ignore_deps=True, ignore_parent=True)
            print("Created image {0}".format(image.image_name))
            if push and image.image_index:
                Syncer().push(image)

@an_action()
def make_pushable(collector, **kwargs):
    """Make only the pushable images and their dependencies"""
    configuration = collector.configuration
    configuration["harpoon"].do_push = True
    configuration["harpoon"].only_pushable = True
    make_all(collector, **kwargs)

@an_action(needs_image=True)
def run(collector, image, **kwargs):
    """Run specified task in this image"""
    image.build_and_run(collector.configuration["images"])

@an_action()
def list_tasks(collector, tasks, **kwargs):
    """List the available_tasks"""
    print("Available tasks to choose from are:")
    print("Use the --task option to choose one")
    print("")
    keygetter = lambda item: item[1].label
    tasks = sorted(tasks.items(), key=keygetter)
    for label, items in itertools.groupby(tasks, keygetter):
        print("--- {0}".format(label))
        print("----{0}".format("-" * len(label)))
        sorted_tasks = sorted(list(items), key=lambda item: len(item[0]))
        max_length = max(len(name) for name, _ in sorted_tasks)
        for key, task in sorted_tasks:
            desc = dedent(task.description or "").strip().split('\n')[0]
            print("\t{0}{1} :-: {2}".format(" " * (max_length-len(key)), key, desc))
        print("")

@an_action()
def delete_untagged(collector, **kwargs):
    """Find the untagged images and remove them"""
    configuration = collector.configuration
    docker_context = configuration["harpoon"].docker_context
    images = docker_context.images()
    found = False
    for image in images:
        if image["RepoTags"] == ["<none>:<none>"]:
            found = True
            image_id = image["Id"]
            log.info("Deleting untagged image\thash=%s", image_id)
            try:
                docker_context.remove_image(image["Id"])
            except DockerAPIError as error:
                log.error("Failed to delete image\thash=%s\terror=%s", image_id, error)

    if not found:
        log.info("Didn't find any untagged images to delete!")

@an_action()
def show(collector, **kwargs):
    """Show what images we have"""
    configuration = collector.configuration
    flat = configuration.get("harpoon.flat", False)
    only_pushable = configuration.get("harpoon.only_pushable", False)

    for index, layer in enumerate(Builder().layered(configuration["images"], only_pushable=only_pushable)):
        if flat:
            for _, image in layer:
                print(image.image_name)
        else:
            print("Layer {0}".format(index))
            for _, image in layer:
                print("    {0}".format(image.display_line()))
            print("")

@an_action()
def show_pushable(collector, **kwargs):
    """Show what images we have"""
    collector.configuration['harpoon'].only_pushable = True
    show(collector, **kwargs)

@an_action(needs_image=True)
def print_dockerfile(collector, image, **kwargs):
    """Print a dockerfile for the specified image"""
    print('\n'.join(image.docker_file.docker_lines))

@an_action()
def print_all_dockerfiles(collector, **kwargs):
    """Print all the dockerfiles"""
    for name, image in collector.configuration["images"].items():
        print("{0}".format(name))
        print("-" * len(name))
        kwargs["image"] = image
        print_dockerfile(collector, **kwargs)

# Make it so future use of @an_action doesn't result in more default tasks
info["is_default"] = False



from harpoon.option_spec.image_objs import Volumes, Context
from harpoon.ship.builders.normal import NormalBuilder
from harpoon.ship.builders.base import BuilderBase
from harpoon.ship.runner import Runner

from docker.errors import APIError as DockerAPIError
from input_algorithms.spec_base import NotSpecified
from contextlib import contextmanager
import logging
import uuid
import sys
import six

log = logging.getLogger("harpoon.ship.builders.persistence")

class PersistenceBuilder(BuilderBase):
    def build(self, conf, context, stream):
        """Do a persistence build!"""
        existing_image = None

        # Find an existing image if it exists
        for image in conf.harpoon.docker_context.images():
            if "{0}:latest".format(conf.image_name) in image["RepoTags"]:
                existing_image = image['Id']

        # If we already have an image, then test to see if any of it's commands
        # have changed. If it's all still cached, then no need to remake the image!
        test_image_name = None
        if existing_image:
            log.info("Building test image for persistence image to see if the cache changed")
            cached, test_image_name = self.make_test_image(conf, context, stream)
            if cached:
                log.info("We have determined no change, so will just use existing image {0}".format(existing_image))
                return True

        try:
            self.make_image(conf, context, stream, existing_image)
        except Exception as error:
            exc_info = sys.exc_info()

            # Make sure if we fail that the test image doesn't
            # make a false positive the next time around
            # because failure means we almost certainly want to
            # go through this dance again
            try:
                if test_image_name:
                    conf.harpoon.docker_context.remove_image(test_image_name)
            except Exception as inner_error:
                log.exception(inner_error)

            six.reraise(*exc_info)

        # Image wasn't cached
        return False

    def make_test_image(self, conf, context, stream):
        docker_file = conf.persistence.make_test_dockerfile(conf.docker_file)
        with self.build_with_altered_context("tester", conf, context, stream, docker_file, tag=True) as (test_conf, cached):
            pass
        return cached, test_conf.image_name

    @contextmanager
    def build_with_altered_context(self, name, conf, context, stream, dockerfile, volumes_from=None, command=None, tag=False):
        conf_image_name = conf.prefixed_image_name
        new_conf = conf.clone()
        if name is not None:
            if tag:
                new_name = "{0}-{1}".format(conf.prefixed_image_name, name)
            else:
                new_name = None
            new_conf.name = name
            new_conf.image_name = new_name
            new_conf.container_id = None
            new_conf.container_name = "{0}-{1}".format(new_name, str(uuid.uuid1())).replace("/", "__")
        else:
            new_name = conf.image_name

        new_conf.bash = NotSpecified
        new_conf.command = NotSpecified

        if command is not None:
            new_conf.bash = command

        # Do we share volumes?
        if volumes_from:
            new_conf.volumes = new_conf.volumes.clone()
            new_conf.volumes.share_with = list(conf.volumes.share_with) + volumes_from

        # Reuse the context or create new one depending on whether context was provided
        if context is not None:
            maker = context.clone_with_new_dockerfile(conf, dockerfile)
        else:
            new_conf.context = Context(enabled=False, parent_dir=new_conf.context.parent_dir)
            maker = new_conf.make_context(docker_file=dockerfile)

        # Create the context manager that removes replaced images
        # If we aren't tagging, then don't do that logic
        # Searching for the images takes time and I want to avoid that
        @contextmanager
        def remover(conf):
            yield
        if new_name is not None:
            remover = self.remove_replaced_images

        # Create our new image!
        with remover(new_conf):
            cached = False
            with maker as new_context:
                cached = NormalBuilder(new_name).build(new_conf, new_context, stream)
                new_conf.image_name = stream.current_container

        yield new_conf, cached

    def run_with_altered_context(self, name, conf, context, stream, dockerfile, volumes_from=None, tag=None, detach=False, command=None):
        """Helper to build and run a new dockerfile"""
        with self.build_with_altered_context(name, conf, context, stream, dockerfile, volumes_from=volumes_from, command=command) as (new_conf, _):
            if detach:
                Runner().run_container(new_conf, {}, detach=True, dependency=True)
            else:
                Runner().run_container(new_conf, {}, detach=False, dependency=False, tag=True)
                new_conf.image_name = new_conf.committed

            return new_conf

    def make_image(self, conf, context, stream, existing_image):
        """
        If the image doesn't already exist, then we just run the normal docker_file
        commands followed by the action and we are done.

        Otherwise, we first create a container with a volume containing the folders from
        the existing container we want to persist. We then make a new image with the normal
        docker_file commands and run it in a container, copy over the folders from the VOLUME
        and commit into an image.

        Finally, we construct an image from that committed image and add a CMD command to
        the one specified in the options, or /bin/bash

        After all this we clean up everything, including that volume we created
        """
        if not existing_image:
            # Don't have an existing image to steal from
            # Just have to make it, no volumes or trickery involved!
            docker_file = conf.persistence.make_first_dockerfile(conf.docker_file)
            with self.build_with_altered_context(None, conf, context, stream, docker_file):
                pass

            # Make the test image so the next time we run this, it's already cached
            self.make_test_image(conf, context, stream)
            return

        # We have an existing image, let's steal from it!
        first_conf = None
        try:
            docker_file = conf.persistence.make_rerunner_prep_dockerfile(conf.docker_file, existing_image)
            first_conf = self.run_with_altered_context("rerunner_prep"
                , conf, context, stream, docker_file, detach=True, command="while true; do sleep 5; done"
                )
            log.info("Built {0}".format(first_conf.image_name))

            # Make the second image, which copies over from the VOLUME into the image
            docker_file = conf.persistence.make_second_dockerfile(conf.docker_file)
            second_image_conf = self.run_with_altered_context("second"
                , conf, context, stream, docker_file, volumes_from=[first_conf.container_id]
                )

            log.info("Built {0}".format(second_image_conf.image_name))

            # Build the final image, which just appends the desired CMD to the end
            docker_file = conf.persistence.make_final_dockerfile(conf.docker_file, second_image_conf.image_name)
            with self.build_with_altered_context(None, conf, None, stream, docker_file):
                pass
        finally:
            if first_conf:
                Runner().stop_container(first_conf, remove_volumes=True)
                try:
                    conf.harpoon.docker_context.remove_image(first_conf.image_name)
                except DockerAPIError as error:
                    log.error(error)




from contextlib import contextmanager
import docker.errors
import humanize
import logging
import os

log = logging.getLogger("harpoon.ship.builders.mixin")

class BuilderBase(object):
    def log_context_size(self, context, conf):
        context_size = humanize.naturalsize(os.stat(context.name).st_size)
        log.info("Building '%s' in '%s' with %s of context", conf.name, conf.context.parent_dir, context_size)

    @contextmanager
    def remove_replaced_images(self, conf):
        current_ids = None
        if not conf.harpoon.keep_replaced:
            try:
                current_id = conf.harpoon.docker_context.inspect_image("{0}:latest".format(conf.image_name))["Id"]
            except docker.errors.APIError as error:
                if str(error).startswith("404 Client Error: Not Found"):
                    current_id = None
                else:
                    raise

        info = {"cached": False}
        yield info

        if current_id and not info.get("cached"):
            log.info("Looking for replaced images to remove")
            untagged = [image["Id"] for image in conf.harpoon.docker_context.images(filters={"dangling": True})]
            if current_id in untagged:
                log.info("Deleting replaced image\ttag=%s\told_hash=%s", "{0}:latest".format(conf.image_name), current_id)
                try:
                    conf.harpoon.docker_context.remove_image(current_id)
                except Exception as error:
                    log.error("Failed to remove replaced image\thash=%s\terror=%s", current_id, error)




from harpoon.errors import BadAmazon
from harpoon import VERSION

from botocore.exceptions import ClientError, NoCredentialsError
from input_algorithms.spec_base import NotSpecified
from six.moves.urllib.parse import urlparse
from contextlib import contextmanager
import logging
import base64
import boto3
import os

log = logging.getLogger("harpoon.amazon.iam")

@contextmanager
def catch_no_credentials(message, **info):
    """Turn a NoCredentialsError into a BadAmazon"""
    try:
        yield
    except NoCredentialsError as error:
        if hasattr(error, "response"):
            info['error_code'] = error.response["ResponseMetadata"]["HTTPStatusCode"]
            info['error_message'] = error.response["Error"]["Message"]
        else:
            info['error_message'] = error.message

        raise BadAmazon(message, **info)

@contextmanager
def catch_boto_400(message, **info):
    """Turn a BotoServerError 400 into a BadAmazon"""
    try:
        yield
    except ClientError as error:
        if str(error.response["ResponseMetadata"]["HTTPStatusCode"]).startswith("4"):
            error_message = error.response["Error"]["Message"]
            raise BadAmazon(message, error_message=error_message, error_code=error.response["ResponseMetadata"]["HTTPStatusCode"], **info)
        else:
            raise

def assume_role(arn):
    log.info("Assuming role as %s", arn)

    session = boto3.session.Session()
    session_name = "harpoon-{0}-".format(VERSION)

    # Clear out empty values
    for name in ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY', 'AWS_SECURITY_TOKEN', 'AWS_SESSION_TOKEN']:
        if name in os.environ and not os.environ[name]:
            del os.environ[name]

    sts = session.client("sts")
    with catch_no_credentials("Couldn't assume role", arn=arn):
        with catch_boto_400("Couldn't assume role", arn=arn):
            creds = sts.assume_role(RoleArn=arn, RoleSessionName=session_name)["Credentials"]

    return boto3.session.Session(
          aws_access_key_id = creds["AccessKeyId"]
        , aws_secret_access_key = creds["SecretAccessKey"]
        , aws_session_token = creds["SessionToken"]
        )

@contextmanager
def assumed_role(arn):
    session = assume_role(arn)

    old_aws_access_key_id = os.environ.get("AWS_ACCESS_KEY_ID", NotSpecified)
    old_aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY", NotSpecified)
    old_aws_security_token = os.environ.get("AWS_SECURITY_TOKEN", NotSpecified)

    try:
        creds = session._session.get_credentials()
        os.environ['AWS_ACCESS_KEY_ID'] = creds.access_key
        os.environ['AWS_SECURITY_TOKEN'] = creds.token
        os.environ['AWS_SECRET_ACCESS_KEY'] = creds.secret_key
        yield
    finally:
        for key, val in (
              ("AWS_ACCESS_KEY_ID", old_aws_access_key_id)
            , ("AWS_SECRET_ACCESS_KEY", old_aws_secret_access_key)
            , ("AWS_SECURITY_TOKEN", old_aws_security_token)
            ):
            if val is not NotSpecified:
                os.environ[key] = val

def get_s3_slip(session, location):
    parsed = urlparse(location)
    log.info("Getting slip from s3://{0}{1}".format(parsed.netloc, parsed.path))
    return session.resource("s3").Object(parsed.netloc, parsed.path[1:]).get()["Body"].read().strip()

def decrypt_kms(session, ciphertext, region):
    return session.client("kms", region).decrypt(CiphertextBlob=base64.b64decode(ciphertext))["Plaintext"]



"""
This is where the mainline sits and is responsible for setting up the logging,
the argument parsing and for starting up Harpoon.
"""

from harpoon.errors import BadDockerConnection
from harpoon.collector import Collector
from harpoon import VERSION

from input_algorithms.spec_base import NotSpecified
from docker.client import Client as DockerClient
from delfick_app import App, DelayedFileType
from docker.utils import kwargs_from_env
import requests
import logging
import os

log = logging.getLogger("harpoon.executor")

def docker_context():
    """Make a docker context"""
    options = kwargs_from_env(assert_hostname=False)
    options["version"] = "auto"
    options["timeout"] = int(os.environ.get("DOCKER_CLIENT_TIMEOUT", 180))

    client = DockerClient(**options)
    try:
        info = client.info()
        log.info("Connected to docker daemon\tdriver=%s\tkernel=%s", info["Driver"], info["KernelVersion"])
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as error:
        raise BadDockerConnection(base_url=options.get('base_url'), error=error)
    return client

class App(App):
    VERSION = VERSION
    cli_categories = ['harpoon']
    cli_description = "Docker client that reads yaml"
    cli_environment_defaults = {"HARPOON_CONFIG": ("--harpoon-config", './harpoon.yml')}
    cli_positional_replacements = [('--task', 'list_tasks'), ('--image', NotSpecified)]

    def execute(self, args_obj, args_dict, extra_args, logging_handler, no_docker=False):
        args_dict["harpoon"]["config"] = args_dict["harpoon"]["config"]()
        args_dict["harpoon"]["extra"] = extra_args

        if not no_docker:
            args_dict["harpoon"]["docker_context"] = docker_context()
        args_dict["harpoon"]["docker_context_maker"] = docker_context

        collector = Collector()
        collector.prepare(args_dict["harpoon"]["config"].name, args_dict)
        if "term_colors" in collector.configuration:
            self.setup_logging_theme(logging_handler, colors=collector.configuration["term_colors"])

        collector.configuration["task_runner"](collector.configuration["harpoon"].chosen_task)

    def setup_other_logging(self, args_obj, verbose=False, silent=False, debug=False):
        logging.getLogger("boto3").setLevel([logging.CRITICAL, logging.ERROR][verbose or debug])
        logging.getLogger("botocore").setLevel([logging.CRITICAL, logging.ERROR][verbose or debug])
        logging.getLogger("requests").setLevel([logging.CRITICAL, logging.ERROR][verbose or debug])

    def specify_other_args(self, parser, defaults):
        parser.add_argument("--harpoon-config"
            , help = "The config file specifying what harpoon should care about"
            , type = DelayedFileType("r")
            , **defaults["--harpoon-config"]
            )

        parser.add_argument("--dry-run"
            , help = "Should Harpoon take any real action or print out what is intends to do"
            , dest = "harpoon_dry_run"
            , action = "store_true"
            )

        parser.add_argument("--task"
            , help = "The task to run"
            , dest = "harpoon_chosen_task"
            , **defaults["--task"]
            )

        parser.add_argument("--image"
            , help = "Specify a particular image"
            , dest = "harpoon_chosen_image"
            , **defaults["--image"]
            )

        parser.add_argument("--assume-role"
            , help = "Role to assume before running the container"
            )

        parser.add_argument("--non-interactive"
            , help = "Make this non interactive"
            , dest = "harpoon_interactive"
            , action = "store_false"
            )

        command = parser.add_mutually_exclusive_group()

        command.add_argument("--command"
            , help = "Specify a command to run for tasks that need one"
            )

        command.add_argument("--bash"
            , help = "Specify a command that will be ran as /bin/bash -c '<command>'"
            )

        parser.add_argument("--silent-build"
            , help = "Make the build process quiet"
            , dest = "harpoon_silent_build"
            , action = "store_true"
            )

        parser.add_argument("--keep-replaced"
            , help = "Don't delete images that have their tag stolen by a new image"
            , dest = "harpoon_keep_replaced"
            , action = "store_true"
            )

        parser.add_argument("--no-intervention"
            , help = "Don't ask to intervene broken builds"
            , dest = "harpoon_no_intervention"
            , action = "store_true"
            )

        parser.add_argument("--intervene-afterwards"
            , help = "Create an intervention container once this container exits"
            , dest = "harpoon_intervene_afterwards"
            , action = "store_true"
            )

        parser.add_argument("--no-cleanup"
            , help = "Don't automatically cleanup after a run"
            , dest = "harpoon_no_cleanup"
            , action = "store_true"
            )

        parser.add_argument("--env"
            , help = "Environment option to start the container with"
            , dest = "extra_env"
            , action = "append"
            )

        parser.add_argument("--port"
            , help = "Specify a port to publish in the running container you make"
            , dest = "extra_ports"
            , action = "append"
            )

        parser.add_argument("--flat"
            , help = "Used with the show command"
            , dest = "harpoon_flat"
            , action = "store_true"
            )

        parser.add_argument("--ignore-missing"
            , help = "Used by the pull commands to ignore if an image doesn't exist"
            , dest = "harpoon_ignore_missing"
            , action = "store_true"
            )

        return parser

main = App.main
if __name__ == '__main__':
    main()



"""
The Builder is responsible for finding and building docker images in
the correct order

Building an image requires building all dependent images, creating the necessary
context, and actually building the current image.
"""

from harpoon.ship.builders.persistence import PersistenceBuilder
from harpoon.ship.builders.squashed import SquashedBuilder
from harpoon.ship.builders.normal import NormalBuilder
from harpoon.ship.builders.base import BuilderBase

from harpoon.errors import NoSuchImage, BadCommand, UserQuit
from harpoon.ship.progress_stream import ProgressStream
from harpoon.option_spec import command_objs
from harpoon.ship.runner import Runner
from harpoon.layers import Layers

from input_algorithms.spec_base import NotSpecified
from itertools import chain
import logging
import six
import sys

log = logging.getLogger("harpoon.ship.builder")

########################
###   PROGRESS STREAM
########################

class BuildProgressStream(ProgressStream):
    def setup(self):
        self.last_line = ""
        self.current_action = ""
        self.current_container = None

    def interpret_line(self, line_detail):
        if "stream" in line_detail:
            self.interpret_stream(line_detail["stream"])
            self.last_line = line_detail["stream"]
        elif "status" in line_detail:
            self.interpret_status(line_detail["status"])
            self.last_line = line_detail["status"]
        else:
            self.interpret_unknown(line_detail)
            self.last_line = str(line_detail)

    def interpret_stream(self, line):
        if line.startswith("Step "):
            action = line[line.find(":")+1:].strip()
            self.current_action = action[:action.find(" ")].strip()

        if line.strip().startswith("---> Running in"):
            self.current_container = line[len("---> Running in "):].strip()
        elif line.strip().startswith("Successfully built"):
            self.current_container = line[len("Successfully built"):].strip()

        if self.last_line.startswith("Step ") and line.strip().startswith("---> "):
            if self.current_action == "FROM":
                self.cached = True
            else:
                self.cached = False

        if line.strip().startswith("---> Running in"):
            self.cached = False
        elif line.strip().startswith("---> Using cache"):
            self.cached = True

        self.add_line(line)

    def interpret_status(self, line):
        if line.startswith("Pulling image"):
            if not line.endswith("\n"):
                line = "{0}\n".format(line)
        else:
            line = "\r{0}".format(line)

        if "already being pulled by another client" in line or "Pulling repository" in line:
            self.cached = False
        self.add_line(line)

########################
###   BUILDER
########################

class Builder(BuilderBase):
    """Build an image from Image configuration"""

    def make_image(self, conf, images, chain=None, parent_chain=None, made=None, ignore_deps=False, ignore_parent=False, pushing=False):
        """Make us an image"""
        made = {} if made is None else made
        chain = [] if chain is None else chain
        parent_chain = [] if parent_chain is None else parent_chain

        if conf.name in made:
            return

        if conf.name in chain and not ignore_deps:
            raise BadCommand("Recursive dependency images", chain=chain + [conf.name])

        if conf.name in parent_chain and not ignore_parent:
            raise BadCommand("Recursive FROM statements", chain=parent_chain + [conf.name])

        if conf.name not in images:
            raise NoSuchImage(looking_for=conf.name, available=images.keys())

        if not ignore_deps:
            for dependency, image in conf.dependency_images():
                self.make_image(images[dependency], images, chain=chain + [conf.name], made=made, pushing=pushing)

        if not ignore_parent:
            parent_image = conf.commands.parent_image
            if not isinstance(parent_image, six.string_types):
                self.make_image(parent_image, images, chain, parent_chain + [conf.name], made=made, pushing=pushing)

        # Should have all our dependencies now
        log.info("Making image for '%s' (%s) - FROM %s", conf.name, conf.image_name, conf.commands.parent_image_name)
        cached = self.build_image(conf, pushing=pushing)
        made[conf.name] = True
        return cached

    def build_image(self, conf, pushing=False):
        """Build this image"""
        with conf.make_context() as context:
            try:
                stream = BuildProgressStream(conf.harpoon.silent_build)
                with self.remove_replaced_images(conf) as info:
                    if conf.persistence is NotSpecified:
                        cached = NormalBuilder().build(conf, context, stream)
                    else:
                        cached = PersistenceBuilder().build(conf, context, stream)
                    info['cached'] = cached
            except (KeyboardInterrupt, Exception) as error:
                exc_info = sys.exc_info()
                if stream.current_container:
                    Runner().stage_build_intervention(conf, stream.current_container)

                if isinstance(error, KeyboardInterrupt):
                    raise UserQuit()
                else:
                    six.reraise(*exc_info)

            try:
                for squash_options, condition in [(conf.squash_after, True), (conf.squash_before_push, pushing)]:
                    if squash_options is not NotSpecified and condition:
                        if type(squash_options) is command_objs.Commands:
                            squash_commands = squash_options.docker_lines_list
                        SquashedBuilder(squash_commands).build(conf, context, stream)
                        cached = False
            except (KeyboardInterrupt, Exception) as error:
                exc_info = sys.exc_info()
                if isinstance(error, KeyboardInterrupt):
                    raise UserQuit()
                else:
                    six.reraise(*exc_info)

        return cached

    def layered(self, images, only_pushable=False):
        """Yield layers of images"""
        if only_pushable:
            operate_on = dict((image, instance) for image, instance in images.items() if instance.image_index)
        else:
            operate_on = images

        layers = Layers(operate_on, all_images=images)
        layers.add_all_to_layers()
        return layers.layered

